#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>
#include <Rmath.h>
#include <omp.h>
#include "pnmath.h"

#ifdef NEED_DUPLICATE_ATTRIB
static void (DUPLICATE_ATTRIB)(SEXP to, SEXP from) {
    SET_ATTRIB(to, duplicate(ATTRIB(from)));
    SET_OBJECT(to, OBJECT(from));
    IS_S4_OBJECT(from) ?  SET_S4_OBJECT(to) : UNSET_S4_OBJECT(to);
}
#endif

#define DUPLICATE_ATTRIB(to, from) do {\
  SEXP __from__ = (from); \
  if (ATTRIB(__from__) != R_NilValue) { \
    SEXP __to__ = (to); \
    (DUPLICATE_ATTRIB)(__to__, __from__);       \
  } \
} while (0)

#define _(String) (String)
#define R_MSG_NA  _("NaNs produced")
#define R_MSG_NONNUM_MATH _("Non-numeric argument to mathematical function")

#define incMod(i,n) (i = (++i == n) ? 0 : i)

/*
  Data base for cutoffs determining the number of threads to use
*/

#define MAX_VNMATH 1000
#define VNNAME_MAX 256
static int nvnmath = 0;
static struct { char name[VNNAME_MAX]; int n2, n4, n8; } vnmath[MAX_VNMATH];

static int findVNmathIndex(const char *name) {
    int i;
    for (i = 0; i < nvnmath; i++)
	if (strncmp(name, vnmath[i].name, VNNAME_MAX) == 0)
	    return i;
    return -1;
}

static int vnmathEnter(const char *name, int n2, int n4, int n8)
{
    int idx = findVNmathIndex(name);
    if (idx < 0) {
	if (nvnmath < MAX_VNMATH) {
	    idx = nvnmath;
	    nvnmath++;
	}
	else error("Table is full.");
    }
    strncpy(vnmath[idx].name, name, VNNAME_MAX);
    vnmath[idx].n2 = n2;
    vnmath[idx].n4 = n4;
    vnmath[idx].n8 = n8;
    return idx;
}

SEXP do_vnmathEnter(SEXP sName, SEXP sN2, SEXP sN4, SEXP sN8)
{
    const char *name;
    int n2, n4, n8, idx;

    if (TYPEOF(sName) != STRSXP || LENGTH(sName) != 1)
	error("bad 'name' argument type");
    name = CHAR(STRING_ELT(sName, 0));
    n2 = asInteger(sN2);
    n4 = asInteger(sN4);
    n8 = asInteger(sN8);
    idx = vnmathEnter(name, n2, n4, n8);
    return ScalarInteger(idx);
}

/*
  Determining the processor count
*/

static int PMAX = 1;

/* For now this uses 1, 2, 4, or 8 processors. */
#define FIXUP_NUM_PROCS do { \
    int idx = -P; \
    if (n > vnmath[idx].n2 && vnmath[idx].n2 > 0 && PMAX >= 2) \
	if (n > vnmath[idx].n4 && vnmath[idx].n4 > 0 && PMAX >= 4) \
	    if (n > vnmath[idx].n8 && vnmath[idx].n8 > 0 && PMAX >= 8) \
		P = 8; \
	    else P = 4; \
	else P = 2; \
    else P = 1; \
    } while (0)


/*
  Math 1 Group
*/

#define MATH1_LOOP_BODY(val, call, arg, naflag) do { \
        if (ISNA (arg)) val = NA_REAL; \
        else if (ISNAN(arg)) val = R_NaN; \
        else { \
            val = (call); \
            if (ISNAN(val)) naflag = 1; \
        } \
    } while (0)


#define MATH1_LOOP(call, naflag) \
    for (i = 0; i < n; i++) { \
	double ai = a[i]; \
        MATH1_LOOP_BODY(y[i], call, ai, naflag); \
    }

static SEXP math1(SEXP sa, double (*f)(double), SEXP lcall, int P)
{
    SEXP sy;
    int i, n, na;
    double *a, *y;
    int naflag;

    if (!isNumeric(sa))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa);

    if (na == 0) return(allocVector(REALSXP, 0));

    n = na;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); y = REAL(sy); naflag = 0;

    if (n == 1)
	MATH1_LOOP_BODY(y[0], f(a[0]), a[0], naflag);
    else {
	if (P <= 0) FIXUP_NUM_PROCS;

	if (P > 1)
	    /* We could use an 'if' clause in the omp pragma intead of
	       one omp branch and one sequential branch but at least for
	       gcc at the time of writing separate code seems to produce
	       better performance */
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) 
	    MATH1_LOOP(f(ai), naflag)
        else
	    MATH1_LOOP(f(ai), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    DUPLICATE_ATTRIB(sy, sa);
    UNPROTECT(2);
    return sy;
} /* math1 */

int PRIMVAL(SEXP);

#ifdef isObject
# undef isObject
#endif
#define isObject(s)   (OBJECT(s) != 0)

#define MIGHT_NEED_DISPATCH(args) \
    ! (args != R_NilValue && ! isObject(CAR(args)) &&		\
       (CDR(args) == R_NilValue || ! isObject(CADR(args))))

#define checkArity(a,b) Rf_checkArityCall(a,b,call)
void Rf_checkArityCall(SEXP, SEXP, SEXP);

#ifdef isComplex
# undef isComplex
#endif
#define isComplex(s) (TYPEOF(s) == CPLXSXP)
static SEXP (*complex_math1)(SEXP, SEXP, SEXP, SEXP);

static SEXP (*do_math1)(SEXP, SEXP, SEXP, SEXP);
static SEXP (*do_math2)(SEXP, SEXP, SEXP, SEXP);
static SEXP (*do_math3)(SEXP, SEXP, SEXP, SEXP);
static SEXP (*do_math4)(SEXP, SEXP, SEXP, SEXP);

#define DEFMATH_V_SETUP(name)  \
    static int idx = -1;       \
    int p = asInteger(sP);     \
    if (p == 0) {					\
	if (idx < 0) idx = findVNmathIndex(#name);	\
	p = -idx;					\
    }

#define DEFMATH_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math1(sa, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (PRIMVAL(op) != 46 && \
            MIGHT_NEED_DISPATCH(args)) \
	    return do_math1(call, op, args, env); \
	if (isComplex(CAR(args))) \
	    return complex_math1(call, op, args, env);\
	sa = CAR(args); \
	return math1(sa, name, R_NilValue, -idx); }

DEFMATH_1(floor)
DEFMATH_1(ceil)
DEFMATH_1(sqrt)
DEFMATH_1(sign)

DEFMATH_1(exp)
DEFMATH_1(expm1)
DEFMATH_1(log1p)
DEFMATH_1(cos)
DEFMATH_1(sin)
DEFMATH_1(tan)
DEFMATH_1(acos)
DEFMATH_1(asin)

DEFMATH_1(cosh)
DEFMATH_1(sinh)
DEFMATH_1(tanh)
DEFMATH_1(acosh)
DEFMATH_1(asinh)
DEFMATH_1(atanh)

DEFMATH_1(lgammafn)
DEFMATH_1(gammafn)

DEFMATH_1(digamma)
DEFMATH_1(trigamma)

DEFMATH_1(gamma_cody)


/*
  Math 2 Group
*/

/* To support recycling we initialize ia, etc, to -1 and declare it
   firtprivate. Then the loop code uses the macro INC_IDX to compute
   the real index as i % na if ia < 0 and otherwise does an increment
   followed by a comparison to the argiment length.  That way the
   modulus computation occurs only once per chunk. */

#define INC_IDX(ix, nx) do { \
        if (nx == n) ix = i; \
        else if (nx == 1) ix = 0; \
        else if(ix < 0) ix = i % nx; \
        else { ix++; if (ix >= nx) ix = 0; } \
    } while (0)

#define MATH2_LOOP_BODY(val, call, arg1, arg2, naflag) do { \
	if (ISNA (arg1) || ISNA (arg2)) val = NA_REAL; \
	else if (ISNAN(arg1) || ISNAN(arg2)) val = R_NaN; \
	else { \
	    val = (call); \
	    if (ISNAN(val)) (naflag) = 1; \
	} \
    } while (0)

#define MATH2_LOOP(call, naflag) \
    for (i = 0; i < n; i++) { \
	double ai, bi;\
        INC_IDX(ia, na); \
	INC_IDX(ib, nb); \
	ai = a[ia]; \
	bi = b[ib]; \
        MATH2_LOOP_BODY(y[i], call, ai, bi, naflag); \
    }

static SEXP math2(SEXP sa, SEXP sb,
		  double (*f)(double, double), SEXP lcall,
		  int P)
{
    SEXP sy;
    int i, n, na, nb;
    int ia = -1, ib = -1;
    double *a, *b, *y;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb);

    if ((na == 0) || (nb == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); y = REAL(sy); naflag = 0;

    if (n == 1)
	MATH2_LOOP_BODY(y[0], f(a[0], b[0]), a[0], b[0], naflag);
    else {
	if (P <= 0) FIXUP_NUM_PROCS;

	if (P > 1)
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) firstprivate(na, nb, ia, ib)
	    MATH2_LOOP(f(ai, bi), naflag)
        else
            MATH2_LOOP(f(ai, bi), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    UNPROTECT(3);
    return sy;
} /* math2 */

static SEXP math2_1(SEXP sa, SEXP sb, SEXP sI,
		    double (*f)(double, double, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, n, na, nb;
    int ia = -1, ib = -1;
    double *a, *b, *y;
    int i_1;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb);

    if ((na == 0) || (nb == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);

    if (n == 1)
	MATH2_LOOP_BODY(y[0], f(a[0], b[0], i_1), a[0], b[0], naflag);
    else {
	if (P <= 0) FIXUP_NUM_PROCS;

	if (P > 1)		    
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) firstprivate(na, nb, ia, ib)
	    MATH2_LOOP(f(ai, bi, i_1), naflag)
        else
            MATH2_LOOP(f(ai, bi, i_1), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    UNPROTECT(3);
    return sy;
} /* math2_1 */

static SEXP math2_2(SEXP sa, SEXP sb, SEXP sI, SEXP sJ,
		    double (*f)(double, double, int, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, n, na, nb;
    int ia = -1, ib = -1;
    double *a, *b, *y;
    int i_1,i_2;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb);

    if ((na == 0) || (nb == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);
    i_2 = asInteger(sJ);

    if (n == 1)
	MATH2_LOOP_BODY(y[0], f(a[0], b[0], i_1, i_2), a[0], b[0], naflag);
    else {
	if (P <= 0) FIXUP_NUM_PROCS;
	if (P > 1)		    
#pragma omp parallel for default(shared) private(i) num_threads(P)\
        reduction(&:naflag) firstprivate(na, nb, ia, ib)
	    MATH2_LOOP(f(ai, bi, i_1, i_2), naflag)
	else
            MATH2_LOOP(f(ai, bi, i_1, i_2), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    UNPROTECT(3);
    return sy;
} /* math2_2 */

static SEXP (*complex_math2)(SEXP, SEXP, SEXP, SEXP);

#define DEFMATH_2(name)	\
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math2(sa, sb, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
        return math2(sa, sb, name, R_NilValue, -idx); }

#define DEFMATH_2_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sI, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math2_1(sa, sb, sI, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sI; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sI = CADDR(args); \
        return math2_1(sa, sb, sI, name, R_NilValue, -idx); }

#define DEFMATH_2_2(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sI, SEXP sJ, SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math2_2(sa, sb, sI, sJ, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sI, sJ;					  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sI = CADDR(args); \
	sJ = CADDDR(args); \
        return math2_2(sa, sb, sI, sJ, name, R_NilValue, -idx); }

DEFMATH_2(lbeta)
DEFMATH_2(beta)
DEFMATH_2(lchoose)
DEFMATH_2(choose)

DEFMATH_2_1(dchisq)
DEFMATH_2_2(pchisq)
DEFMATH_2_2(qchisq)

DEFMATH_2_1(dexp)
DEFMATH_2_2(pexp)
DEFMATH_2_2(qexp)

DEFMATH_2_1(dgeom)
DEFMATH_2_2(pgeom)
DEFMATH_2_2(qgeom)

DEFMATH_2_1(dpois)
DEFMATH_2_2(ppois)
DEFMATH_2_2(qpois)

DEFMATH_2_1(dt)
DEFMATH_2_2(pt)
DEFMATH_2_2(qt)

#ifdef NOT_THREAD_SAFE
DEFMATH_2_1(dsignrank)
DEFMATH_2_2(psignrank)
DEFMATH_2_2(qsignrank)

DEFMATH_2(bessel_j)
DEFMATH_2(bessel_y)
#endif
DEFMATH_2(psigamma)


/*
  Math 3 Group
*/

#define MATH3_LOOP_BODY(val, call, arg1, arg2, arg3, naflag) do { \
	if (ISNA (arg1) || ISNA (arg2)|| ISNA (arg3)) val = NA_REAL; \
	else if (ISNAN(arg1) || ISNAN(arg2)|| ISNAN(arg3)) val = R_NaN; \
	else { \
	    val = (call); \
	    if (ISNAN(val)) (naflag) = 1; \
	} \
    } while (0)

#define MATH3_LOOP(call, naflag) \
    for (i = 0; i < n; i++) { \
	double ai, bi, ci; \
        INC_IDX(ia, na); \
	INC_IDX(ib, nb); \
        INC_IDX(ic, nc); \
	ai = a[ia]; \
	bi = b[ib]; \
	ci = c[ic]; \
        MATH3_LOOP_BODY(y[i], call, ai, bi, ci, naflag); \
    }

static SEXP math3_1(SEXP sa, SEXP sb, SEXP sc, SEXP sI,
		    double (*f)(double, double, double, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, n, na, nb, nc;
    int ia = -1, ib = -1, ic = -1;
    double *a, *b, *c, *y;
    int i_1;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc);

    if ((na == 0) || (nb == 0) || (nc == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);

    if (n == 1)
	MATH3_LOOP_BODY(y[0], f(a[0], b[0], c[0], i_1),
			a[0], b[0], c[0], naflag);
    else {
	if (P <= 0) FIXUP_NUM_PROCS;

	if (P > 1)
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) firstprivate(na, nb, nc, ia, ib, ic)
	    MATH3_LOOP(f(ai, bi, ci, i_1), naflag)
	else
	    MATH3_LOOP(f(ai, bi, ci, i_1), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    UNPROTECT(4);
    return sy;
} /* math3_1 */

static SEXP math3_2(SEXP sa, SEXP sb, SEXP sc, SEXP sI, SEXP sJ,
		    double (*f)(double, double, double, int, int), SEXP lcall,
		    int P)
{
    SEXP sy;
    int i, n, na, nb, nc;
    int ia = -1, ib = -1, ic = -1;
    double *a, *b, *c, *y;
    int i_1,i_2;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc);

    if ((na == 0) || (nb == 0) || (nc == 0)) return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); y = REAL(sy); naflag = 0;
    i_1 = asInteger(sI);
    i_2 = asInteger(sJ);

    if (P <= 0) FIXUP_NUM_PROCS;

    if (n == 1)
	MATH3_LOOP_BODY(y[0], f(a[0], b[0], c[0], i_1, i_2),
			a[0], b[0], c[0], naflag);
    else {
	if (P > 1)
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) firstprivate(na, nb, nc, ia, ib, ic)
	    MATH3_LOOP(f(ai, bi, ci, i_1, i_2), naflag)
	else
	    MATH3_LOOP(f(ai, bi, ci, i_1, i_2), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    UNPROTECT(4);
    return sy;
} /* math3_2 */

#define DEFMATH_3_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sI, SEXP sP) {	\
	DEFMATH_V_SETUP(name) \
        return math3_1(sa, sb, sc, sI, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sI;					  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sI = CADDDR(args); \
        return math3_1(sa, sb, sc, sI, name, R_NilValue, -idx); }
#define DEFMATH_3_2(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sI, SEXP sJ, \
		       SEXP sP) {				    \
	DEFMATH_V_SETUP(name) \
        return math3_2(sa, sb, sc, sI, sJ, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sI, sJ;					  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sI = CADDDR(args); \
	sJ = CAD4R(args); \
        return math3_2(sa, sb, sc, sI, sJ, name, R_NilValue, -idx); }

DEFMATH_3_1(dbeta)
DEFMATH_3_2(pbeta)
DEFMATH_3_2(qbeta)

DEFMATH_3_1(dbinom)
DEFMATH_3_2(pbinom)
DEFMATH_3_2(qbinom)

DEFMATH_3_1(dcauchy)
DEFMATH_3_2(pcauchy)
DEFMATH_3_2(qcauchy)

DEFMATH_3_1(df)
DEFMATH_3_2(pf)
DEFMATH_3_2(qf)

DEFMATH_3_1(dgamma)
DEFMATH_3_2(pgamma)
DEFMATH_3_2(qgamma)

DEFMATH_3_1(dlnorm)
DEFMATH_3_2(plnorm)
DEFMATH_3_2(qlnorm)

DEFMATH_3_1(dlogis)
DEFMATH_3_2(plogis)
DEFMATH_3_2(qlogis)

DEFMATH_3_1(dnbinom)
DEFMATH_3_2(pnbinom)
DEFMATH_3_2(qnbinom)

DEFMATH_3_1(dnorm)
DEFMATH_3_2(pnorm)
DEFMATH_3_2(qnorm)

DEFMATH_3_1(dunif)
DEFMATH_3_2(punif)
DEFMATH_3_2(qunif)

DEFMATH_3_1(dweibull)
DEFMATH_3_2(pweibull)
DEFMATH_3_2(qweibull)

DEFMATH_3_1(dnchisq)
DEFMATH_3_2(pnchisq)
DEFMATH_3_2(qnchisq)

DEFMATH_3_1(dnt)
DEFMATH_3_2(pnt)
DEFMATH_3_2(qnt)

#ifdef NOT_THREAD_SAFE
DEFMATH_3_1(dwilcox)
DEFMATH_3_2(pwilcox)
DEFMATH_3_2(qwilcox)

DEFMATH3(bessel_i)
DEFMATH3(bessel_k)
#endif

/*
  Math 4 Group
*/

#define MATH4_LOOP_BODY(val, call, arg1, arg2, arg3, arg4, naflag) do { \
	if (ISNA(arg1) || ISNA(arg2)|| ISNA(arg3) || ISNA(arg4)) \
            val = NA_REAL; \
	else if (ISNAN(arg1) || ISNAN(arg2)|| ISNAN(arg3) || ISNAN(arg4)) \
            val = R_NaN; \
	else { \
	    val = (call); \
	    if (ISNAN(val)) (naflag) = 1; \
	} \
    } while (0)

#define MATH4_LOOP(call, naflag) \
    for (i = 0; i < n; i++) { \
	double ai, bi, ci, di; \
        INC_IDX(ia, na); \
	INC_IDX(ib, nb); \
        INC_IDX(ic, nc); \
        INC_IDX(id, nd); \
	ai = a[ia]; \
	bi = b[ib]; \
	ci = c[ic]; \
	di = d[id]; \
        MATH4_LOOP_BODY(y[i], call, ai, bi, ci, di, naflag); \
    }

static SEXP math4_1(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI,
		    double (*f)(double, double, double, double, int),
		    SEXP lcall, int P)
{
    SEXP sy;
    int i, n, na, nb, nc, nd;
    int ia = -1, ib = -1, ic = -1, id = -1;
    double *a, *b, *c, *d, *y;
    int i_1;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc) || !isNumeric(sd))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc); nd = LENGTH(sd);

    if ((na == 0) || (nb == 0) || (nc == 0) || (nd == 0))
	return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc; if (n < nd) n = nd;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sd = coerceVector(sd, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); d = REAL(sd); y = REAL(sy);
    naflag = 0;
    i_1 = asInteger(sI);

    if (n == 1)
	MATH4_LOOP_BODY(y[0], f(a[0], b[0], c[0], d[0], i_1),
			a[0], b[0], c[0], d[0], naflag);
    else {
	if (P <= 0) FIXUP_NUM_PROCS;
		
	if (P > 1)
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) firstprivate(na, nb, nc, nd, ia, ib, ic, id)
	    MATH4_LOOP(f(ai, bi, ci, di, i_1), naflag)
	else
	    MATH4_LOOP(f(ai, bi, ci, di, i_1), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    else if (n == nd) { DUPLICATE_ATTRIB(sy, sd); }
    UNPROTECT(5);
    return sy;
} /* math4_1 */

static SEXP math4_2(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI, SEXP sJ,
		    double (*f)(double, double, double, double, int, int),
		    SEXP lcall, int P)
{
    SEXP sy;
    int i, n, na, nb, nc, nd;
    int ia = -1, ib = -1, ic = -1, id = -1;
    double *a, *b, *c, *d, *y;
    int i_1,i_2;
    int naflag;

    if (!isNumeric(sa) || !isNumeric(sb) || !isNumeric(sc) || !isNumeric(sd))
	errorcall(lcall, R_MSG_NONNUM_MATH);

    na = LENGTH(sa); nb = LENGTH(sb); nc = LENGTH(sc); nd = LENGTH(sd);

    if ((na == 0) || (nb == 0) || (nc == 0) || (nd == 0))
	return(allocVector(REALSXP, 0));

    n = na; if (n < nb) n = nb; if (n < nc) n = nc; if (n < nd) n = nd;

    PROTECT(sa = coerceVector(sa, REALSXP));
    PROTECT(sb = coerceVector(sb, REALSXP));
    PROTECT(sc = coerceVector(sc, REALSXP));
    PROTECT(sd = coerceVector(sd, REALSXP));
    PROTECT(sy = allocVector(REALSXP, n));

    a = REAL(sa); b = REAL(sb); c = REAL(sc); d = REAL(sd); y = REAL(sy);
    naflag = 0;
    i_1 = asInteger(sI);
    i_2 = asInteger(sJ);

    if (P <= 0) FIXUP_NUM_PROCS;
		
    if (n == 1)
	MATH4_LOOP_BODY(y[0], f(a[0], b[0], c[0], d[0], i_1, i_2),
			a[0], b[0], c[0], d[0], naflag);
    else {
	if (P > 1)
#pragma omp parallel for default(shared) private(i) num_threads(P) \
        reduction(&:naflag) firstprivate(na, nb, nc, nd, ia, ib, ic, id)
	    MATH4_LOOP(f(ai, bi, ci, di, i_1, i_2), naflag)
	else
	    MATH4_LOOP(f(ai, bi, ci, di, i_1, i_2), naflag)
    }

    if(naflag) warningcall(lcall, R_MSG_NA);
    if (n == na) { DUPLICATE_ATTRIB(sy, sa); }
    else if (n == nb) { DUPLICATE_ATTRIB(sy, sb); }
    else if (n == nc) { DUPLICATE_ATTRIB(sy, sc); }
    else if (n == nd) { DUPLICATE_ATTRIB(sy, sd); }
    UNPROTECT(5);
    return sy;
} /* math4_2 */

#define CAD5R(e)CAR(CDR(CDR(CDR(CDR(CDR(e))))))

#define DEFMATH_4_1(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI, \
		       SEXP sP) {				    \
	DEFMATH_V_SETUP(name) \
        return math4_1(sa, sb, sc, sd, sI, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sd, sI; \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sd = CADDDR(args); \
	sI = CAD4R(args); \
        return math4_1(sa, sb, sc, sd, sI, name, R_NilValue, -idx); }
#define DEFMATH_4_2(name) \
    SEXP do_##name##_v(SEXP sa, SEXP sb, SEXP sc, SEXP sd, SEXP sI, SEXP sJ, \
		       SEXP sP) { \
	DEFMATH_V_SETUP(name) \
        return math4_2(sa, sb, sc, sd, sI, sJ, name, R_NilValue, p); } \
    SEXP do_##name##_i(SEXP call, SEXP op, SEXP args, SEXP env) { \
	SEXP sa, sb, sc, sd, sI, sJ;				  \
        static int idx = -1; \
	if (idx < 0) idx = findVNmathIndex(#name); \
	checkArity(op, args); \
	if (isComplex(CAR(args)) || \
            (PRIMVAL(op) == 0 && isComplex(CADR(args)))) \
            return complex_math2(call, op, args, env); \
	sa = CAR(args); \
	sb = CADR(args); \
	sc = CADDR(args); \
	sd = CADDDR(args); \
	sI = CAD4R(args); \
	sJ = CAD5R(args); \
        return math4_2(sa, sb, sc, sd, sI, sJ, name, R_NilValue, -idx); }

DEFMATH_4_1(dhyper)
DEFMATH_4_2(phyper)
DEFMATH_4_2(qhyper)

DEFMATH_4_1(dnbeta)
DEFMATH_4_2(pnbeta)
DEFMATH_4_2(qnbeta)

DEFMATH_4_1(dnf)
DEFMATH_4_2(pnf)
DEFMATH_4_2(qnf)

#ifdef UNIMP
DEFMATH_4_1(dtukey)
#endif
DEFMATH_4_2(ptukey)
DEFMATH_4_2(qtukey)


/*
  Replacing Builtin Functions.
 */

typedef SEXP (*CCODE)();
CCODE (PRIMFUN)(SEXP);
void (SET_PRIMFUN)(SEXP, CCODE);

CCODE R_get_arith_function(int);

static int getNumCPU(void);

SEXP do_initRpartest()
{
    PMAX = getNumCPU();
    if (PMAX < 0) PMAX = 1;

    do_math1 = R_get_arith_function(1);
    do_math2 = R_get_arith_function(2);
    do_math3 = R_get_arith_function(3);
    do_math4 = R_get_arith_function(4);
    complex_math1 = R_get_arith_function(11);
    complex_math2 = R_get_arith_function(12);
    return R_NilValue;
}

/* It ought to be sufficient to call omp_get_num_procs here and be
   done with it.  But on but Mac OS X box with 8 cores this returns 1
   if OMP_NUM_THREADS is not set and the value of OMP_NUM_THREADS if
   it is set. So it's back to using this hacked code for now. */

#include <unistd.h> /* for sysconf stuff */
#include <sys/types.h>
#include <sys/sysctl.h> /* for sysctl stuff */

static int getNumCPU(void)
{
    int numCPU = -1;
#if defined(HW_NCPU) /* BSD, e.g. Mac OS X */
    int mib[2];
    size_t len;

    mib[0] = CTL_HW;
    mib[1] = HW_NCPU;
    len = sizeof(numCPU);
    sysctl(mib, 2, &numCPU, &len, NULL, 0);
#elif defined(_SC_NPROCESSORS_CONF) /* Linux; maybe some others */
    numCPU = sysconf(_SC_NPROCESSORS_CONF);
#else
    numCPU = omp_get_num_procs();
#endif
    return numCPU;
}

SEXP do_NumProcs()
{
    return ScalarInteger(getNumCPU());
}

SEXP do_GetPMAX(void)
{
    return ScalarInteger(PMAX);
}

#define PMAX_SANITY_LIMIT 1024

SEXP do_SetPMAX(SEXP sP)
{
    int P = asInteger(sP);
    if (P > 0 && P < PMAX_SANITY_LIMIT)
	PMAX = P;
    return ScalarInteger(PMAX);
}

static void replaceBuiltin(const char *name, CCODE oldfun, CCODE newfun)
{
    SEXP v, symbol = install(name);
    
    if ((v = INTERNAL(symbol)) != R_NilValue && TYPEOF(v) == BUILTINSXP
	&& PRIMFUN(v) == oldfun)
	SET_PRIMFUN(v, newfun);
    if ((v = SYMVALUE(symbol)) != R_UnboundValue && TYPEOF(v) == BUILTINSXP
	&& PRIMFUN(v) == oldfun)
	SET_PRIMFUN(v, newfun);
}
    
SEXP do_replaceBuiltins(void)
{
    /* Math 1 Group */
    replaceBuiltin("exp", do_math1, do_exp_i);

    /*replaceBuiltin("floor", do_math1, do_floor_i);
      replaceBuiltin("ceil", do_math1, do_ceil_i);*/
    replaceBuiltin("sqrt", do_math1, do_sqrt_i);
    /*replaceBuiltin("sign", do_math1, do_sign_i);*/

    replaceBuiltin("exp", do_math1, do_exp_i);
    replaceBuiltin("expm1", do_math1, do_expm1_i);
    replaceBuiltin("log1p", do_math1, do_log1p_i);
    replaceBuiltin("cos", do_math1, do_cos_i);
    replaceBuiltin("sin", do_math1, do_sin_i);
    replaceBuiltin("tan", do_math1, do_tan_i);
    replaceBuiltin("acos", do_math1, do_acos_i);
    replaceBuiltin("asin", do_math1, do_asin_i);

    replaceBuiltin("cosh", do_math1, do_cosh_i);
    replaceBuiltin("sinh", do_math1, do_sinh_i);
    replaceBuiltin("tanh", do_math1, do_tanh_i);
    replaceBuiltin("acosh", do_math1, do_acosh_i);
    replaceBuiltin("asinh", do_math1, do_asinh_i);
    replaceBuiltin("atanh", do_math1, do_atanh_i);

    replaceBuiltin("lgammafn", do_math1, do_lgammafn_i);
    replaceBuiltin("gammafn", do_math1, do_gammafn_i);

    replaceBuiltin("digamma", do_math1, do_digamma_i);
    replaceBuiltin("trigamma", do_math1, do_trigamma_i);

    replaceBuiltin("gamma_cody", do_math1, do_gamma_cody_i);


    /* Math 2 Group */
    replaceBuiltin("lbeta", do_math2, do_lbeta_i);
    replaceBuiltin("beta", do_math2, do_beta_i);
    replaceBuiltin("lchoose", do_math2, do_lchoose_i);
    replaceBuiltin("choose", do_math2, do_choose_i);

    replaceBuiltin("dchisq", do_math2, do_dchisq_i);
    replaceBuiltin("pchisq", do_math2, do_pchisq_i);
    replaceBuiltin("qchisq", do_math2, do_qchisq_i);

    replaceBuiltin("dexp", do_math2, do_dexp_i);
    replaceBuiltin("pexp", do_math2, do_pexp_i);
    replaceBuiltin("qexp", do_math2, do_qexp_i);

    replaceBuiltin("dgeom", do_math2, do_dgeom_i);
    replaceBuiltin("pgeom", do_math2, do_pgeom_i);
    replaceBuiltin("qgeom", do_math2, do_qgeom_i);

    replaceBuiltin("dpois", do_math2, do_dpois_i);
    replaceBuiltin("ppois", do_math2, do_ppois_i);
    replaceBuiltin("qpois", do_math2, do_qpois_i);

    replaceBuiltin("dt", do_math2, do_dt_i);
    replaceBuiltin("pt", do_math2, do_pt_i);
    replaceBuiltin("qt", do_math2, do_qt_i);

#ifdef NOT_THREAD_SAFE
    replaceBuiltin("dsignrank", do_math2, do_dsignrank_i);
    replaceBuiltin("psignrank", do_math2, do_psignrank_i);
    replaceBuiltin("qsignrank", do_math2, do_qsignrank_i);

    replaceBuiltin("bessel_j", do_math2, do_bessel_j_i);
    replaceBuiltin("bessel_y", do_math2, do_bessel_y_i);
#endif
    replaceBuiltin("psigamma", do_math2, do_psigamma_i);


    /* Math 3 Group */
    replaceBuiltin("dbeta", do_math3, do_dbeta_i);
    replaceBuiltin("pbeta", do_math3, do_pbeta_i);
    replaceBuiltin("qbeta", do_math3, do_qbeta_i);

    replaceBuiltin("dbinom", do_math3, do_dbinom_i);
    replaceBuiltin("pbinom", do_math3, do_pbinom_i);
    replaceBuiltin("qbinom", do_math3, do_qbinom_i);

    replaceBuiltin("dcauchy", do_math3, do_dcauchy_i);
    replaceBuiltin("pcauchy", do_math3, do_pcauchy_i);
    replaceBuiltin("qcauchy", do_math3, do_qcauchy_i);

    replaceBuiltin("df", do_math3, do_df_i);
    replaceBuiltin("pf", do_math3, do_pf_i);
    replaceBuiltin("qf", do_math3, do_qf_i);

    replaceBuiltin("dgamma", do_math3, do_dgamma_i);
    replaceBuiltin("pgamma", do_math3, do_pgamma_i);
    replaceBuiltin("qgamma", do_math3, do_qgamma_i);

    replaceBuiltin("dlnorm", do_math3, do_dlnorm_i);
    replaceBuiltin("plnorm", do_math3, do_plnorm_i);
    replaceBuiltin("qlnorm", do_math3, do_qlnorm_i);

    replaceBuiltin("dlogis", do_math3, do_dlogis_i);
    replaceBuiltin("plogis", do_math3, do_plogis_i);
    replaceBuiltin("qlogis", do_math3, do_qlogis_i);

    replaceBuiltin("dnbinom", do_math3, do_dnbinom_i);
    replaceBuiltin("pnbinom", do_math3, do_pnbinom_i);
    replaceBuiltin("qnbinom", do_math3, do_qnbinom_i);

    replaceBuiltin("dnorm", do_math3, do_dnorm_i);
    replaceBuiltin("pnorm", do_math3, do_pnorm_i);
    replaceBuiltin("qnorm", do_math3, do_qnorm_i);

    replaceBuiltin("dunif", do_math3, do_dunif_i);
    replaceBuiltin("punif", do_math3, do_punif_i);
    replaceBuiltin("qunif", do_math3, do_qunif_i);

    replaceBuiltin("dweibull", do_math3, do_dweibull_i);
    replaceBuiltin("pweibull", do_math3, do_pweibull_i);
    replaceBuiltin("qweibull", do_math3, do_qweibull_i);

    replaceBuiltin("dnchisq", do_math3, do_dnchisq_i);
    replaceBuiltin("pnchisq", do_math3, do_pnchisq_i);
    replaceBuiltin("qnchisq", do_math3, do_qnchisq_i);

    replaceBuiltin("dnt", do_math3, do_dnt_i);
    replaceBuiltin("pnt", do_math3, do_pnt_i);
    replaceBuiltin("qnt", do_math3, do_qnt_i);

#ifdef NOT_THREAD_SAFE
    replaceBuiltin("dwilcox", do_math3, do_dwilcox_i);
    replaceBuiltin("pwilcox", do_math3, do_pwilcox_i);
    replaceBuiltin("qwilcox", do_math3, do_qwilcox_i);

    replaceBuiltin("bessel_i", do_math3, do_bessel_i_i);
    replaceBuiltin("bessel_k", do_math3, do_bessel_k_i);
#endif


    /* Math 4 Group */
    replaceBuiltin("dhyper", do_math4, do_dhyper_i);
    replaceBuiltin("phyper", do_math4, do_phyper_i);
    replaceBuiltin("qhyper", do_math4, do_qhyper_i);

    replaceBuiltin("dnbeta", do_math4, do_dnbeta_i);
    replaceBuiltin("pnbeta", do_math4, do_pnbeta_i);
    replaceBuiltin("qnbeta", do_math4, do_qnbeta_i);

    replaceBuiltin("dnf", do_math4, do_dnf_i);
    replaceBuiltin("pnf", do_math4, do_pnf_i);
    replaceBuiltin("qnf", do_math4, do_qnf_i);

#ifdef UNIMP
    replaceBuiltin("dtukey", do_math4, do_dtukey_i);
#endif
    replaceBuiltin("ptukey", do_math4, do_ptukey_i);
    replaceBuiltin("qtukey", do_math4, do_qtukey_i);

    return R_NilValue;
}
